/*
 * VARiD: A Variation Detection Framework for Colorspace and Letterspace platforms
 * Contact: <varid at cs dot toronto dot edu>
 * Copyright 2010 the authors (see README).
 *
 * This file is part of VARiD.
 *
 * VARiD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * VARiD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VARiD.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include <math.h>
#include "va_assert.h"
#include "va_log.h"
#include "va_time.h"
#include "varid.h"

/******************************************************************************
* VARID_COVMAPS computes the letter and colorspace coverage maps
*   Here we compute a few coverage-related stats. We compute the overall
*   coverage, the number of times we see each color and each letter, and
*   finally the emission support for a color (col_em) and against a
*   color (noncol_em), and the same for letters. More details to come.
*
*
*
*
*   GEN_SIZE, READSTARTS, READS
*
*
*   note: stats CUTOFFs, size CUTOFFs
* TODO: quality values
* !!! Asuuming T start when we do first letter translation???
*******************************************************************************/

extern double*  cs_error_cnt;
extern double*  cs_error_dpts;
extern unsigned int cs_error_size;
extern double*  ls_error_cnt;
extern double*  ls_error_dpts;
extern unsigned int ls_error_size;


static int update_map_with_read(VA_VecU8 seq, VA_VecU8 qvals,
				VA_UINT32 start, VA_DOUBLE * qval_errors,
				VA_DOUBLE * counts, VA_DOUBLE * em, VA_DOUBLE * non_em, VA_INT32 * coverage, 
				VA_UINT32 ref_size , double gap_error, int haplome, VA_INT32 * qval_array, VA_VecU8 * qval_ordering, int unique_qvals ) {
	VA_ASSERT(seq.size==qvals.size);
	int j; int used_positions=0;
	for (j=0; j<seq.size; j++) {
		//If this is a funky letter or we have fallen off the END!
		if ( seq.data[j] == 5 || j + start > ref_size ) {
			continue;
		}
		coverage[ (haplome - 1) * ref_size + (start + j - 1) ] ++;
		//otherwise lets use this information in the covmap
		VA_DOUBLE error;
		if ( seq.data[j]==4) {
			error = gap_error;
		} else {
			error = qval_errors[qvals.data[j]];
			qval_array[(start+j-1)*(5*unique_qvals)+unique_qvals*seq.data[j]+qval_ordering->data[qvals.data[j]]]++;	
		}
		                   //row number     row size          column
		//int matrix_index = (seq.data[j]) * ref_size + (start + j -1) ;
		int matrix_index = (start+j-1)*5 + seq.data[j];
		VA_ASSERT(matrix_index< ref_size*5);
		//update covmap counts
		counts[matrix_index] ++;
		//update covmap yes / non probabilities
		VA_ASSERT(error >= 0.0 && error <= 1.0);
		em[matrix_index] += log(1-error);
		non_em[matrix_index] += log(error/3);
		used_positions++;		
	}
	return used_positions;
}


Covmaps_t varid_covmaps(
    CData_t cdata,
    VA_MatI32 * letter_qvals,
    VA_MatI32 * color_qvals,
    VA_VecU8 * qval_ordering,
    int unique_qvals)
{
    struct timespec ts;
    struct timespec te;
    struct timespec tdiff;
    int rc;
    Covmaps_t covmap;
    int usedreads = 0; /* the number of reads used */
    int i;
    size_t sz = cdata.ref.size;
    size_t nrreads = cdata.reads.size;
    ReadsNode_t * rd = (ReadsNode_t *)cdata.reads.data;
    int zeroindexskippedreads = 0;

    rc = va_GetTime(&ts);
    VA_ASSERT(rc == 0);

    /* coverage maps of various kind */
    /* the amount of coverage over a certain position from each allele */
    covmap.coverage = va_allocMatI32(sz, 2);  
    memset(covmap.coverage.data, 0, sizeof(VA_INT32) * 
           covmap.coverage.size_x * covmap.coverage.size_y);

    /* color emissions, in log space */
    /* histogram of colors */
    covmap.colors = va_allocMatDouble(sz, 5);  
    memset(covmap.colors.data, 0, sizeof(VA_DOUBLE) * 
           covmap.colors.size_x * covmap.colors.size_y);

    /* col_emission */
    covmap.col_em = va_allocMatDouble(sz, 5);  
    memset(covmap.col_em.data, 0, sizeof(VA_DOUBLE) * 
           covmap.col_em.size_x * covmap.col_em.size_y);

    /* noncol_emission */
    covmap.noncol_em = va_allocMatDouble(sz, 5);  
    memset(covmap.noncol_em.data, 0, sizeof(VA_DOUBLE) * 
           covmap.noncol_em.size_x * covmap.noncol_em.size_y);

    /* letter emissions, in log space */
    /* histogram of letters */
    covmap.letters = va_allocMatDouble(sz, 5);  
    memset(covmap.letters.data, 0, sizeof(VA_DOUBLE) * 
           covmap.letters.size_x * covmap.letters.size_y);

    /* let_emission */
    covmap.let_em = va_allocMatDouble(sz, 5);  
    memset(covmap.let_em.data, 0, sizeof(VA_DOUBLE) * 
           covmap.let_em.size_x * covmap.let_em.size_y);

    /* nonlet_emission */
    covmap.nonlet_em = va_allocMatDouble(sz, 5);  
    memset(covmap.nonlet_em.data, 0, sizeof(VA_DOUBLE) * 
           covmap.nonlet_em.size_x * covmap.nonlet_em.size_y);

    for (i = 0; i < nrreads; i++) {
        int haplome;
        VA_DOUBLE * error_mat;

        /* assign the haplome */
        if (cdata.src == SRC_REAL) {
            /* is real data, we don't know the haplome, so we work with 1 by
             * default
             */
            haplome = 1;
        } else {
            haplome = rd[i].extras.haplome;
        }

        //isForward = (rd[i].strand == '+') ? TRUE : FALSE;
	if ( rd[i].isLetterspace ) {
		error_mat = cdata.error_mat_ls;
	} else {
		error_mat = cdata.error_mat_cs;
	}
		
	int used_positions = 0;
	
	/*if ( (rd[i].ls_seq.size!=1 || rd[i].cs_seq.size=0) && (rd[i].ls_seq.size==0 || rd[i].cs_seq.size!=0) ) {
		fprintf(stderr,"Read is neither color space or letter space? Offending read name: %s\n",rd[i].name);
		fprintf(stderr,"ls_seq_size %d, cs_seq_size %d\n",rd[i].ls_seq.size,rd[i].cs_seq.size);
		exit(1);
	}*/
	VA_ASSERT(rd[i].ls_seq.size == rd[i].ls_qvals.size);
	used_positions += update_map_with_read(rd[i].ls_seq, rd[i].ls_qvals,
				rd[i].ls_chrstart, error_mat,
				covmap.letters.data, covmap.let_em.data, covmap.nonlet_em.data, covmap.coverage.data, 
				cdata.ref.size, gap_error, haplome, letter_qvals->data, qval_ordering, unique_qvals );
	VA_ASSERT(rd[i].cs_seq.size == rd[i].cs_qvals.size);
	used_positions += update_map_with_read(rd[i].cs_seq, rd[i].cs_qvals,
				rd[i].cs_chrstart, error_mat,
				covmap.colors.data, covmap.col_em.data, covmap.noncol_em.data, covmap.coverage.data,
				cdata.ref.size, gap_error, haplome, color_qvals->data, qval_ordering, unique_qvals );

        if (used_positions == 0) {
            zeroindexskippedreads++;
        } else {
            usedreads++;
	}
    }

    /* print time information */
    rc = va_GetTime(&te);
    VA_ASSERT(rc == 0);

    tdiff = va_TimeSub(&te, &ts);
    fprintf(stderr, "\t covmaps built in: %ld.%03ld secs, "
        "Used reads in the covmaps: %d/%lu, Skipped for 0 index:%d\n",
        tdiff.tv_sec, tdiff.tv_nsec/NS_IN_MS,
        usedreads, cdata.reads.size, zeroindexskippedreads);

    /*FILE* letters_file = fopen("letters.txt","w");
    //print_genome(letters_file,cdata.ref);
    print_matrix_double_raw(letters_file,covmap.letters.data, sz, 5);
    FILE* colors_file = fopen("colors.txt","w");
    //print_genome(colors_file,cdata.ref);
    print_matrix_double_raw(colors_file,covmap.colors.data, sz, 5);
    FILE* colem_file = fopen("colem.txt","w");
    //print_genome(colem_file,cdata.ref);
    print_matrix_double_raw(colem_file,covmap.col_em.data, sz, 5);
    FILE* letem_file = fopen("letem.txt","w");
    //print_genome(letem_file,cdata.ref);
    print_matrix_double_raw(letem_file,covmap.let_em.data, sz, 5);
    FILE* non_colem_file = fopen("noncol_em.txt","w");
    //print_genome(non_colem_file,cdata.ref);
    print_matrix_double_raw(non_colem_file,covmap.noncol_em.data, sz, 5);
    FILE* non_letem_file = fopen("nonlet_em.txt","w");
    //print_genome(non_letem_file,cdata.ref);
    print_matrix_double_raw(non_letem_file,covmap.nonlet_em.data, sz, 5);
    FILE* coverage_file = fopen("coverage.txt","w");
    //print_genome(coverage_file,cdata.ref);
    print_matrix_int_raw(coverage_file,covmap.coverage.data, sz, 2);*/
    
    

    return (covmap);
} 


