/*
 * VARiD: A Variation Detection Framework for Colorspace and Letterspace platforms
 * Contact: <varid at cs dot toronto dot edu>
 * Copyright 2010 the authors (see README).
 *
 * This file is part of VARiD.
 *
 * VARiD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * VARiD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VARiD.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "va_log.h"
#include "va_assert.h"
#include "varid.h"

/******************************************************************************
* COL2NUC translates a color-space sequence into letter-space sequence
*   nuc = COL2NUC(startletter, colorref, out_type) translates a color-space
*   (cs) sequence into a letter-space sequence. The cs sequence has to be
*   in 0-4 mode, with 0-3 being the first 4 colors, and 4 representing a
*   gap. The startletter can be in either 0-4 space (a,c,g,t,-) or actual
*   <lowercase> letters a,c,g,t,-.
*
*   Mandatory parameters:
*   startletter - the first letter in the sequence (or last Letter in the
*       linker). Can be in 0-4 space (a vector) or a,c,g,t,- (a string)
*       space.
*   colorref - the color reference vector, in 0-4 space.
*
*   Optional parameters:
*   out_type - the type of nucleotide output. can be 'ns' (number-space) or
*   'ls' (letter-space). If this is not supplied, the output will be in the
*   same type as the given startletter.
*
*   Returns:
*   the nucleotide translation of the color reference starting with the
*   given startletter.
*
*   See also NUC2COL, NUC2NUC, GET_COMPLEMENT
*
*   Author Adrian V Dalca
*******************************************************************************/

VA_VecU8 col2nuc(
    VA_UINT8    startletter, 
    VA_VecU8 colorref)
{
    int i;
    VA_VecU8 nuc;
    VA_VecU8 ret;
    int out_type;

    VA_ASSERT((colorref.data != NULL) && (colorref.size > 0));

    /* determine output type : ls or ns */
    out_type = isalpha(startletter) ? 'l' : 'n';
        
    /* Work in ns: */
    if (isalpha(startletter))
    {
        VA_VecU8 startlet_vec;

        startlet_vec.data = &startletter;
        startlet_vec.size = 1;
        startlet_vec = nuc2nuc(startlet_vec);
        startletter = startlet_vec.data[0];
        free(startlet_vec.data);
    }

    nuc.size = colorref.size + 1;
    nuc.data = (VA_UINT8 *)malloc(sizeof(VA_UINT8) * nuc.size);
    VA_ASSERT(nuc.data != NULL);
    memset(nuc.data, 0, sizeof(VA_UINT8) * nuc.size);

    /* work in 0-3 space, with bitwise xor... */
    nuc.data[0] = startletter; 
    VA_UINT8 last_letter = startletter;
    for (i = 0; i < colorref.size; i++) {
        if (colorref.data[i] == 5) {
            nuc.data[i + 1] = 5;
	} else if (colorref.data[i]==4) {
	    nuc.data[i+1]=4;
        } else {
            last_letter = nuc.data[i + 1] = last_letter ^ colorref.data[i]; 
	}
    }

    /* get the right output type */
    if (out_type == 'l')
    {
        ret = nuc2nuc(nuc);
        free(nuc.data);
    }
    else
    {
        ret = nuc;
    }

    return (ret);
}    


        
