/*
 * VARiD: A Variation Detection Framework for Colorspace and Letterspace platforms
 * Contact: <varid at cs dot toronto dot edu>
 * Copyright 2010 the authors (see README).
 *
 * This file is part of VARiD.
 *
 * VARiD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * VARiD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VARiD.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include "va_assert.h"
#include "va_log.h"
#include "va_time.h"
#include "varid.h"
#include "config.h"

static char ** initGroupFiles(
    int          nfiles);

static VA_Vector getFastaReference(void);

static int cmpSortI32(
   const void * rec1,
   const void * rec2);

char ** filenames = NULL;


/*******************************************************************************
* VARID_SINGLE prepare the data and run the VARiD algorithm.
*   varid_single(readsfile, reffile, don_name, thr_vec_p, path) will run
*   the varid algorithm after preparing the data. If the reference file
*   (mfa) will have more than one reference, the readsfile will be split
*   based on the references they are aligned to, in the folder <path>, and
*   the reads will be loaded piecewise based on the reference that is being
*   processed. The result will be outputed in prettybase format.
*
*   parameters:
*       readfile - path to the file of reads, which has to be in prettyprint
*           format for now...
*       reffile - path to the mfa reference file.
*       don_name - the name of the donor genome, to be included in the output
*       thr_vec_p - the probability threshold for p
*       path - the path for storing the remporary split of reads.
*
*   Author: Adrian V. Dalca
*   (adalca at mit dot edu)
*******************************************************************************/
void varid_single(){
    SnpGroups_t snpgroup;
    VA_UINT32 i;
    VA_UINT32 j;
    VA_UINT32 gnr;
    struct timespec mastertic;
    struct timespec readtic;
    struct timespec te;
    struct timespec tdiff;
    int rc;
    Fasta_t * fasta;
    VA_Vector fastaVec;
    Maps_t    maps;
    SnpGroup_t * p;
    VA_INT32 * fid_reads = NULL;
    VA_VecDouble   thr_vec_snp;



    /* give some information to the user */
    fprintf(stderr, "\n\t - alignments: %s \n\t - ref-file:   %s\n\n"
        "Detected: \n", alignments_filename, reference_filename);

    /* prepare master clock */
    rc = va_GetTime(&mastertic);
    VA_ASSERT(rc == 0);

    /* threshold vectors */
    thr_vec_snp = va_allocVecDouble(1);
    thr_vec_snp.data[0] = 0.0;
       
    /* get the reference and shift snpgroups */
    fastaVec = getFastaReference();
    VA_ASSERT(fastaVec.size > 0);

    /* get the reads */
    rc = va_GetTime(&readtic);
    VA_ASSERT(rc == 0);
 
    fprintf(stderr, "Splitting reads\n");

    filenames = initGroupFiles(fastaVec.size);

    snpgroup.size = fastaVec.size;
    snpgroup.data = (SnpGroup_t *)malloc(snpgroup.size * sizeof(SnpGroup_t));
    VA_ASSERT(snpgroup.data != NULL);
    memset(snpgroup.data, 0, sizeof(SnpGroup_t));
    p = (SnpGroup_t *)snpgroup.data;

    fasta = (Fasta_t *)fastaVec.data;

    for (i = 0; i < fastaVec.size; i++) {
        p[i].range = va_allocVecI32(2);
        p[i].range.data[0] = 1;
        p[i].range.data[1] = fasta[i].seq.size;
        p[i].refstart = 1;
        p[i].chr = va_allocVecU8(strlen(fasta[i].hdr));
        memcpy(p[i].chr.data, fasta[i].hdr, p[i].chr.size);
    }
            
    split_reads(filenames, &snpgroup, &fid_reads, fastaVec);

    rc = va_GetTime(&te);
    VA_ASSERT(rc == 0);
    tdiff = va_TimeSub(&te, &readtic);
    /* done read prep */
    fprintf(stderr, "\t Done: in: %ld.%03ld secs\n",
     	tdiff.tv_sec, tdiff.tv_nsec / NS_IN_MS);



    /* pre-process data and run VARiD on each SNP group */

    /* run processing and VARiD for every snp group */
    for (gnr = 0; gnr < snpgroup.size; gnr++) {
        if (filenames[gnr] != NULL) {
            VA_VecU8 g_ref;
            VA_Vector g_reads;
            VA_VecI32 g_readstarts;
            VA_VecI32 g_ins_vec; 
            CData_t cdata;
            VA_MatDouble p;
            VA_MatDouble tot;

            fprintf(stderr, "------------------------------\n");
	    fprintf(stderr, "Processing fasta entry : %s - %lubp\n", fasta[gnr].hdr,fasta[gnr].ref.size);
            fprintf(stderr, "Processing group number : %d\n", gnr + 1);

	    rc = va_GetTime(&readtic);
	    VA_ASSERT(rc == 0);

            fprintf(stderr, "Loading reads ...\n");
	    VA_VecU8 qval_ordering = va_allocVecU8(255);
	    memset(qval_ordering.data,0,sizeof(VA_UINT8)*255);
            get_reads(filenames[gnr], fid_reads[gnr],  
                      &g_reads, &g_ins_vec, &fasta[gnr], &qval_ordering);

	    int i; int temp=0;
	    for (i=0; i<qval_ordering.size; i++) {
		temp+=qval_ordering.data[i];
	    } 
	    temp=0;
	    for (i=0; i<qval_ordering.size; i++) {
		if (qval_ordering.data[i]==1) {
			qval_ordering.data[i]=temp++;
		}
	    } 
	    

	    rc = va_GetTime(&te);
	    VA_ASSERT(rc == 0);
	    tdiff = va_TimeSub(&te, &readtic);
	    fprintf(stderr, "\t Done in %ld.%03ld secs\n", 
	       tdiff.tv_sec, tdiff.tv_nsec / NS_IN_MS);


            /* form the readstarts vector */
            g_readstarts = va_allocVecI32(g_reads.size);

            /* make sure there are reads in this snp group */
            /*if (g_reads.size == 0){
                fprintf(stderr, "Group %d has no reads\n", gnr + 1);
                continue;
            }*/
      
            /* get the reads-generated insertion vector */
            //if (g_ins_vec.size == 0){
            //    g_ins_vec = va_allocVecI32(1);
            //    g_ins_vec.data[0] = 1;
            //}
            
            /* prepare the insertion vector */
            qsort(g_ins_vec.data, g_ins_vec.size, sizeof(VA_INT32), cmpSortI32);
            
	    //print out the g_ins_vec 
	    //printf("g_ins_vec:\n");
	    //for (i=0; i<g_ins_vec.size; i++) {
	    //printf("%d,",g_ins_vec.data[i]);
	    //}
	    //printf("\n");
            /* extract needed reference and insert */
            g_ref = va_allocVecU8(fasta[gnr].ref.size + g_ins_vec.size);
	    int ref_index=0; int inserts_index=0;
	    for (i=0; i<g_ref.size; i++) {
		if (inserts_index<g_ins_vec.size && g_ins_vec.data[inserts_index] == i-inserts_index) {
			g_ref.data[i]=4;
			inserts_index++;
		} else {
			g_ref.data[i]=fasta[gnr].ref.data[ref_index++];
		}
	    }

	   // print out the reference 
	    /*VA_VecU8 g_ref_ls = nuc2nuc(g_ref);
	    printf("The reference in memory\n");
	    for(i=0; i<g_ref_ls.size; i++) {
	    printf("%c",g_ref_ls.data[i]);
	    } 
	    printf("\n");*/

            /* align reads */
            g_reads = align_reads(g_ref, g_reads, g_ins_vec);

            ReadsNode_t * rd = (ReadsNode_t *)g_reads.data;
            for (j = 0; j < g_reads.size; j++) {
               g_readstarts.data[j] = MIN(rd[j].ls_chrstart,rd[j].cs_chrstart);
			VA_ASSERT(rd[j].ls_seq.size == rd[j].ls_qvals.size);	
		}

            /* create and populate the cdata structure */
            cdata = real_cdata_prep(TRUE);
            cdata.ref = g_ref;
            cdata.reads = g_reads;
            cdata.readstarts = g_readstarts;
            //cdata.read_size = 35;

            /* run the VARiD Algorithm and get initial predictions */

            /* varid algorithm */
            varid(cdata, 0, &maps, &snpgroup.data[gnr].covmap, &p, &qval_ordering);

	    /*{
		FILE * f = fopen("p_matrix_before.txt","w");
		print_genome(f,g_ref);
		print_matrix_double(f,p.data,p.size_x,p.size_y);
		fclose(f);
	    }*/
            predict(p, maps, &snpgroup.data[gnr].maxvals_2, 
                    &snpgroup.data[gnr].lmat, 
                    &snpgroup.data[gnr].cmat);
	    free(maps.symm.data);
	    /*{
		FILE * f = fopen("p_matrix_after.txt","w");
		print_genome(f,g_ref);
		print_matrix_double(f,p.data,p.size_x,p.size_y);
		fclose(f);
	    }*/

            if (p.data != NULL)
                free(p.data);

 
            /* Fix SNPs TODO: use sparse?
             * destroy_state doesn'yt use thresholds anymore...
             */
            destroy_state(cdata, snpgroup.data[gnr].cmat, g_reads, 
                          g_readstarts, snpgroup.data[gnr].covmap, thr_vec_snp, 
                          snpgroup.data[gnr].lmat, &snpgroup.data[gnr].ch_list, 
                          &tot, &snpgroup.data[gnr].top);
	    /*{
	    int y;
	    for (y=0; y< snpgroup.data[gnr].lmat.size_y; y++ ){
		int x;
		for (x=0; x< snpgroup.data[gnr].lmat.size_x; x++) {
			printf("%d",snpgroup.data[gnr].lmat.data[y*snpgroup.data[gnr].lmat.size_x + x]);
		}
		printf("\n");
	    }
	    }*/
            snpgroup.data[gnr].ref = g_ref;
            snpgroup.data[gnr].readstarts = g_readstarts;

            if (maps.full_letter.data != NULL) {
                free(maps.full_letter.data);
                maps.full_letter.data = NULL;
            }

            if (maps.call_letter.data != NULL) {
                free(maps.call_letter.data);
                maps.call_letter.data = NULL;
            }

            if (maps.call_color.data != NULL) {
                free(maps.call_color.data);
                maps.call_color.data = NULL;
            }

            rd  = (ReadsNode_t *)cdata.reads.data;
            for (j = 0; j < cdata.reads.size; j++) {
                if (rd[j].name != NULL)
                    free(rd[j].name);
                if (rd[j].ls_seq.data != NULL)
                    free(rd[j].ls_seq.data);
                if (rd[j].ls_qvals.data != NULL)
                    free(rd[j].ls_qvals.data);
                if (rd[j].cs_seq.data != NULL)
                    free(rd[j].cs_seq.data);
                if (rd[j].cs_qvals.data != NULL)
                    free(rd[j].cs_qvals.data);
                if (rd[j].ins.data != NULL)
                    free(rd[j].ins.data);
            }
            free(rd);

            if (cdata.readstarts.data != NULL)
                free(cdata.readstarts.data);

            {
                VA_VecU8 hap;
                VA_VecU8 pred1_whole;
                VA_VecU8 pred2_whole;

                memset(&hap, 0, sizeof(hap));


                get_pred(snpgroup.data[gnr].covmap, maps,
                         snpgroup.data[gnr].ch_list,
                         snpgroup.data[gnr].maxvals_2, snpgroup.data[gnr].lmat,
                         hap, hap, &pred1_whole, &pred2_whole);

		
	    /*{
	    int y;
	    for (y=0; y< snpgroup.data[gnr].ch_list.size_y; y++ ){
		int x;
		for (x=0; x< snpgroup.data[gnr].ch_list.size_x; x++) {
			printf("%d",snpgroup.data[gnr].ch_list.data[y*snpgroup.data[gnr].ch_list.size_x + x]);
		}
		printf("\n");
	    }
	    printf("Pred1\n");
	    for (y=0; y<pred1_whole.size; y++) {
		printf("%d",pred1_whole.data[y]);
	    }
	    printf("\nPred2\n");
	    for (y=0; y<pred2_whole.size; y++) {
		printf("%d",pred2_whole.data[y]);
	    }
	    printf("\n");
	    }*/

                writePrettybase(snpgroup.data[gnr].range.data[0],
                            fasta[gnr].hdr,  pred1_whole, pred2_whole,
                            snpgroup.data[gnr].ref, snpgroup.data[gnr].maxvals_2);
		free(pred1_whole.data);
		free(pred2_whole.data);
            }
	    free(snpgroup.data[gnr].lmat.data);
	    free(snpgroup.data[gnr].cmat.data);
	    free(snpgroup.data[gnr].ch_list.data);
	    free(snpgroup.data[gnr].covmap.coverage.data);
	    free(snpgroup.data[gnr].covmap.colors.data);
	    free(snpgroup.data[gnr].covmap.letters.data);
	    free(snpgroup.data[gnr].covmap.col_em.data);
	    free(snpgroup.data[gnr].covmap.noncol_em.data);
	    free(snpgroup.data[gnr].covmap.let_em.data);
	    free(snpgroup.data[gnr].covmap.nonlet_em.data);
	    free(snpgroup.data[gnr].range.data);
	    free(snpgroup.data[gnr].chr.data);	   
	    free(snpgroup.data[gnr].maxvals_2.data);
	    free(g_ref.data);
        }
    }
    free(fid_reads);
    for (i=0; i<fastaVec.size; i++) {
	if (fasta[i].seq.size>0) {
		free (fasta[i].seq.data);
	}
	if (fasta[i].ref.size>0){ 
		free(fasta[i].ref.data);
	}
	free(fasta[i].hdr);
    }
    free(fasta);
    free(thr_vec_snp.data);
    free(snpgroup.data);
	    rc = va_GetTime(&te);
	    VA_ASSERT(rc == 0);

	    tdiff = va_TimeSub(&te, &mastertic);
	    fprintf(stderr, "Time for computation: %ld.%03ld secs\n",
		    tdiff.tv_sec, tdiff.tv_nsec/NS_IN_MS);
}

static char ** initGroupFiles(
    int          nfiles)
{
    int i;
    char ** filenames;
    char buf[BUFSIZ];


    VA_ASSERT(nfiles > 0);

    filenames = (char **)malloc(sizeof(char *) * nfiles);
    VA_ASSERT(filenames != NULL);


    for (i = 0; i < nfiles; i++)
    {
        char * p = va_basename(alignments_filename);

        buf[0] = 0;
        sprintf(buf, "%s/%s.group%d", temp_directory, p, i + 1);
        free(p);
        filenames[i] = (char *)malloc(strlen(buf) + 1);
        VA_ASSERT(filenames[i] != NULL);
    
        strcpy(filenames[i], buf);
    }

    return (filenames);
}

static VA_Vector getFastaReference(void) {
    VA_Vector fasta;
    int i;
    int j;
    Fasta_t * p;
    int ret;


    ret = va_readfasta(&fasta);
    VA_ASSERT(ret == 0);

    p = (Fasta_t *)fasta.data;
    for (i = 0; i < fasta.size; i++) {
        VA_VecU8 v;

        p[i].ref = va_allocVecU8(p[i].seq.size - 1);

        for (j = 0; j < p[i].ref.size; j++)
            p[i].ref.data[j] = tolower(p[i].seq.data[j]);
        v = nuc2nuc(p[i].ref);
        free(p[i].ref.data);
        memcpy(&p[i].ref, &v, sizeof(v));
        fprintf(stderr, "\t %20s:%lu nt \n", p[i].hdr, p[i].ref.size);
	if (predict_start>0) {
		if (hmm_end>p[i].ref.size) {
			hmm_end=p[i].ref.size;
		}
		fprintf(stderr,"hmm_start : %lld, hmm_end : %lld, predict_start : %lld, predict_end : %lld \n",
			hmm_start, hmm_end, predict_start, predict_end);
		VA_ASSERT(hmm_start>0 && hmm_start < hmm_end && hmm_end <=p[i].ref.size);
		VA_VecU8 temp =va_allocVecU8(hmm_end-hmm_start+1);
		memcpy(temp.data,p[i].ref.data+hmm_start-1,sizeof(VA_UINT8)*(hmm_end-hmm_start+1));
		free(p[i].ref.data);
		p[i].ref=temp;
		temp =va_allocVecU8(hmm_end-hmm_start+1+1); //need to end with newline!
		memcpy(temp.data,p[i].seq.data+hmm_start-1,sizeof(VA_UINT8)*(hmm_end-hmm_start+1));
		temp.data[temp.size-1]='\n';
		free(p[i].seq.data);
		p[i].seq=temp;
	}
    }

    return (fasta);
}

static int cmpSortI32(
   const void * rec1,
   const void * rec2)
{
   VA_INT32 p1;
   VA_INT32 p2;

   VA_ASSERT((rec1 != NULL) && (rec2 != NULL));

   p1 = *(VA_INT32 *)rec1;
   p2 = *(VA_INT32 *)rec2;

   return (p1 - p2);
}


